@@ -3,7 +3,7 @@ require "ostruct"
 module ActiveRecord::Import::ConnectionAdapters ; end
 
 module ActiveRecord::Import #:nodoc:
-  class Result < Struct.new(:failed_instances, :num_inserts)
+  class Result < Struct.new(:failed_instances, :num_inserts, :ids)
   end
 
   module ImportSupport #:nodoc:
@@ -25,6 +25,71 @@ module ActiveRecord::Import #:nodoc:
   end
 end
 
+class ActiveRecord::Associations::CollectionProxy
+  def import(*args, &block)
+    @association.import(*args, &block)
+  end
+end
+
+class ActiveRecord::Associations::CollectionAssociation
+  def import(*args, &block)
+    unless owner.persisted?
+      raise ActiveRecord::RecordNotSaved, "You cannot call import unless the parent is saved"
+    end
+
+    options = args.last.is_a?(Hash) ? args.pop : {}
+
+    model_klass = self.reflection.klass
+    symbolized_foreign_key = self.reflection.foreign_key.to_sym
+    symbolized_column_names = model_klass.column_names.map(&:to_sym)
+
+    owner_primary_key = self.owner.class.primary_key
+    owner_primary_key_value = self.owner.send(owner_primary_key)
+
+    # assume array of model objects
+    if args.last.is_a?( Array ) and args.last.first.is_a? ActiveRecord::Base
+      if args.length == 2
+        models = args.last
+        column_names = args.first
+      else
+        models = args.first
+        column_names = symbolized_column_names
+      end
+
+      if !symbolized_column_names.include?(symbolized_foreign_key)
+        column_names << symbolized_foreign_key
+      end
+
+      models.each do |m|
+        m.send "#{symbolized_foreign_key}=", owner_primary_key_value
+      end
+
+      return model_klass.import column_names, models, options
+
+    # supports empty array
+    elsif args.last.is_a?( Array ) and args.last.empty?
+      return ActiveRecord::Import::Result.new([], 0, []) if args.last.empty?
+
+    # supports 2-element array and array
+    elsif args.size == 2 and args.first.is_a?( Array ) and args.last.is_a?( Array )
+      column_names, array_of_attributes = args
+      symbolized_column_names = column_names.map(&:to_s)
+
+      if !symbolized_column_names.include?(symbolized_foreign_key)
+        column_names << symbolized_foreign_key
+        array_of_attributes.each { |attrs| attrs << owner_primary_key_value }
+      else
+        index = symbolized_column_names.index(symbolized_foreign_key)
+        array_of_attributes.each { |attrs| attrs[index] = owner_primary_key_value }
+      end
+
+      return model_klass.import column_names, array_of_attributes, options
+    else
+      raise ArgumentError.new( "Invalid arguments!" )
+    end
+  end
+end
+
 class ActiveRecord::Base
   class << self
 
@@ -44,18 +109,21 @@ class ActiveRecord::Base
     # Returns true if the current database connection adapter
     # supports import functionality, otherwise returns false.
     def supports_import?(*args)
-      connection.supports_import?(*args)
-    rescue NoMethodError
-      false
+      connection.respond_to?(:supports_import?) && connection.supports_import?(*args)
     end
 
     # Returns true if the current database connection adapter
     # supports on duplicate key update functionality, otherwise
     # returns false.
     def supports_on_duplicate_key_update?
-      connection.supports_on_duplicate_key_update?
-    rescue NoMethodError
-      false
+      connection.respond_to?(:supports_on_duplicate_key_update?) && connection.supports_on_duplicate_key_update?
+    end
+
+    # returns true if the current database connection adapter
+    # supports setting the primary key of bulk imported models, otherwise
+    # returns false
+    def support_setting_primary_key_of_imported_objects?
+      connection.respond_to?(:support_setting_primary_key_of_imported_objects?) && connection.support_setting_primary_key_of_imported_objects?
     end
 
     # Imports a collection of values to the database.
@@ -107,6 +175,9 @@ class ActiveRecord::Base
     #   existing model instances in memory with updates from the import.
     # * +timestamps+ - true|false, tells import to not add timestamps \
     #   (if false) even if record timestamps is disabled in ActiveRecord::Base
+    # * +recursive - true|false, tells import to import all autosave association
+    #   if the adapter supports setting the primary keys of the newly imported
+    #   objects.
     #
     # == Examples
     #  class BlogPost < ActiveRecord::Base ; end
@@ -120,7 +191,7 @@ class ActiveRecord::Base
     #  # Example using column_names and array_of_values
     #  columns = [ :author_name, :title ]
     #  values = [ [ 'zdennis', 'test post' ], [ 'jdoe', 'another test post' ] ]
-    #  BlogPost.import columns, values 
+    #  BlogPost.import columns, values
     #
     #  # Example using column_names, array_of_value and options
     #  columns = [ :author_name, :title ]
@@ -142,7 +213,7 @@ class ActiveRecord::Base
     #
     # == On Duplicate Key Update (MySQL only)
     #
-    # The :on_duplicate_key_update option can be either an Array or a Hash. 
+    # The :on_duplicate_key_update option can be either an Array or a Hash.
     #
     # ==== Using an Array
     #
@@ -165,11 +236,25 @@ class ActiveRecord::Base
     # This returns an object which responds to +failed_instances+ and +num_inserts+.
     # * failed_instances - an array of objects that fails validation and were not committed to the database. An empty array if no validation is performed.
     # * num_inserts - the number of insert statements it took to import the data
-    def import( *args )
-      options = { :validate=>true, :timestamps=>true }
+    # * ids - the priamry keys of the imported ids, if the adpater supports it, otherwise and empty array.
+    def import(*args)
+      if args.first.is_a?( Array ) and args.first.first.is_a? ActiveRecord::Base
+        options = {}
+        options.merge!( args.pop ) if args.last.is_a?(Hash)
+
+        models = args.first
+        import_helper(models, options)
+      else
+        import_helper(*args)
+      end
+    end
+
+    def import_helper( *args )
+      options = { :validate=>true, :timestamps=>true, :primary_key=>primary_key }
       options.merge!( args.pop ) if args.last.is_a? Hash
 
-      is_validating = options.delete( :validate )
+      is_validating = options[:validate]
+      is_validating = true unless options[:validate_with_context].nil?
 
       # assume array of model objects
       if args.last.is_a?( Array ) and args.last.first.is_a? ActiveRecord::Base
@@ -191,7 +276,7 @@ class ActiveRecord::Base
         end
         # supports empty array
       elsif args.last.is_a?( Array ) and args.last.empty?
-        return ActiveRecord::Import::Result.new([], 0) if args.last.empty?
+        return ActiveRecord::Import::Result.new([], 0, []) if args.last.empty?
         # supports 2-element array and array
       elsif args.size == 2 and args.first.is_a?( Array ) and args.last.is_a?( Array )
         column_names, array_of_attributes = args
@@ -218,16 +303,26 @@ class ActiveRecord::Base
       return_obj = if is_validating
         import_with_validations( column_names, array_of_attributes, options )
       else
-        num_inserts = import_without_validations_or_callbacks( column_names, array_of_attributes, options )
-        ActiveRecord::Import::Result.new([], num_inserts)
+        (num_inserts, ids) = import_without_validations_or_callbacks( column_names, array_of_attributes, options )
+        ActiveRecord::Import::Result.new([], num_inserts, ids)
       end
 
       if options[:synchronize]
         sync_keys = options[:synchronize_keys] || [self.primary_key]
         synchronize( options[:synchronize], sync_keys)
       end
-
       return_obj.num_inserts = 0 if return_obj.num_inserts.nil?
+
+      # if we have ids, then set the id on the models and mark the models as clean.
+      if support_setting_primary_key_of_imported_objects?
+        set_ids_and_mark_clean(models, return_obj)
+
+        # if there are auto-save associations on the models we imported that are new, import them as well
+        if options[:recursive]
+          import_associations(models, options)
+        end
+      end
+
       return_obj
     end
 
@@ -254,19 +349,19 @@ class ActiveRecord::Base
         instance = new do |model|
           hsh.each_pair{ |k,v| model.send("#{k}=", v) }
         end
-        if not instance.valid?
+        if not instance.valid?(options[:validate_with_context])
           array_of_attributes[ i ] = nil
           failed_instances << instance
         end
       end
       array_of_attributes.compact!
 
-      num_inserts = if array_of_attributes.empty? || options[:all_or_none] && failed_instances.any?
-                      0
+      (num_inserts, ids) = if array_of_attributes.empty? || options[:all_or_none] && failed_instances.any?
+                      [0,[]]
                     else
                       import_without_validations_or_callbacks( column_names, array_of_attributes, options )
                     end
-      ActiveRecord::Import::Result.new(failed_instances, num_inserts)
+      ActiveRecord::Import::Result.new(failed_instances, num_inserts, ids)
     end
 
     # Imports the passed in +column_names+ and +array_of_attributes+
@@ -298,6 +393,7 @@ class ActiveRecord::Base
       columns_sql = "(#{column_names.map{|name| connection.quote_column_name(name) }.join(',')})"
       insert_sql = "INSERT #{options[:ignore] ? 'IGNORE ':''}INTO #{quoted_table_name} #{columns_sql} VALUES "
       values_sql = values_sql_for_columns_and_attributes(columns, array_of_attributes)
+      ids = []
       if not supports_import?
         number_inserted = 0
         values_sql.each do |values|
@@ -309,15 +405,60 @@ class ActiveRecord::Base
         post_sql_statements = connection.post_sql_statements( quoted_table_name, options )
 
         # perform the inserts
-        number_inserted = connection.insert_many( [ insert_sql, post_sql_statements ].flatten,
+        (number_inserted,ids) = connection.insert_many( [ insert_sql, post_sql_statements ].flatten,
                                                   values_sql,
                                                   "#{self.class.name} Create Many Without Validations Or Callbacks" )
       end
-      number_inserted
+      [number_inserted, ids]
     end
 
     private
 
+    def set_ids_and_mark_clean(models, import_result)
+      unless models.nil?
+        import_result.ids.each_with_index do |id, index|
+          models[index].id = id.to_i
+          models[index].instance_variable_get(:@changed_attributes).clear # mark the model as saved
+        end
+      end
+    end
+
+    def import_associations(models, options)
+      # now, for all the dirty associations, collect them into a new set of models, then recurse.
+      # notes:
+      #    does not handle associations that reference themselves
+      #    assumes that the only associations to be saved are marked with :autosave
+      #    should probably take a hash to associations to follow.
+      associated_objects_by_class={}
+      models.each {|model| find_associated_objects_for_import(associated_objects_by_class, model) }
+
+      associated_objects_by_class.each_pair do |class_name, associations|
+        associations.each_pair do |association_name, associated_records|
+          associated_records.first.class.import(associated_records, options) unless associated_records.empty?
+        end
+      end
+    end
+
+    # We are eventually going to call Class.import <objects> so we build up a hash
+    # of class => objects to import.
+    def find_associated_objects_for_import(associated_objects_by_class, model)
+      associated_objects_by_class[model.class.name]||={}
+
+      model.class.reflect_on_all_autosave_associations.each do |association_reflection|
+        associated_objects_by_class[model.class.name][association_reflection.name]||=[]
+
+        association = model.association(association_reflection.name)
+        association.loaded!
+
+        changed_objects = association.select {|a| a.new_record? || a.changed?}
+        changed_objects.each do |child|
+          child.send("#{association_reflection.foreign_key}=", model.id)
+        end
+        associated_objects_by_class[model.class.name][association_reflection.name].concat changed_objects
+      end
+      associated_objects_by_class
+    end
+
     # Returns SQL the VALUES for an INSERT statement given the passed in +columns+
     # and +array_of_attributes+.
     def values_sql_for_columns_and_attributes(columns, array_of_attributes)   # :nodoc:
@@ -331,11 +472,11 @@ class ActiveRecord::Base
           # be sure to query sequence_name *last*, only if cheaper tests fail, because it's costly
           if val.nil? && column.name == primary_key && !sequence_name.blank?
              connection_memo.next_value_for_sequence(sequence_name)
-          else
-            if serialized_attributes.include?(column.name)
-              connection_memo.quote(serialized_attributes[column.name].dump(val), column)
+          elsif column
+            if column.respond_to?(:type_cast_from_user)                         # Rails 4.2 and higher
+              connection_memo.quote(column.type_cast_from_user(val), column)
             else
-              connection_memo.quote(val, column)
+              connection_memo.quote(column.type_cast(val), column)              # Rails 3.1, 3.2, and 4.1
             end
           end
         end
@@ -347,7 +488,7 @@ class ActiveRecord::Base
       AREXT_RAILS_COLUMNS[:create].each_pair do |key, blk|
         if self.column_names.include?(key)
           value = blk.call
-          if index=column_names.index(key)
+          if index=column_names.index(key) || index=column_names.index(key.to_sym)
             # replace every instance of the array of attributes with our value
             array_of_attributes.each{ |arr| arr[index] = value if arr[index].nil? }
           else
@@ -360,7 +501,7 @@ class ActiveRecord::Base
       AREXT_RAILS_COLUMNS[:update].each_pair do |key, blk|
         if self.column_names.include?(key)
           value = blk.call
-          if index=column_names.index(key)
+          if index=column_names.index(key) || index=column_names.index(key.to_sym)
              # replace every instance of the array of attributes with our value
              array_of_attributes.each{ |arr| arr[index] = value }
           else
